#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include<time.h>


#define INF 9999
#define MAX_NOEUD 6

/* Structure de file et fonctions utiles */
typedef struct element {
    int valeur;
    struct element *suivant;
} element;


typedef struct fifo{
  element *debut;
  element *fin;
  int taille;
} fifo;

int is_empty_fifo(fifo *F){
    if (F->taille == 0 && F->debut == NULL && F->fin == NULL)
        return 1;
    else
        return 0;
}

fifo * new_fifo(){
   fifo *F = malloc(sizeof(*F));
   if (F == NULL)
        printf("Erreur d'allocation.");
   F->taille = 0;
   F->debut = NULL;
   F->fin = NULL;
   return F;
}

void push(fifo *F, int s){
    if (F == NULL)
        printf ("La file n'existe pas - On ne peut empiler.");
    element *e = malloc(sizeof(*e));
    if (e == NULL)
        printf("Erreur d'allocation");
    e->valeur = s;
    e->suivant = NULL;
    if (F->taille == 0){
        F->debut = e;
        F->fin = e;
    }
    else{
        F->fin->suivant = e;
        F->fin = e;
    }
    F->taille+=1;
}

void defiler(fifo *F) {
    element *e = malloc(sizeof(*e));
    if (F == NULL || F->taille == 0)
        printf("Erreur dépilement.");
    e = F->debut;
    if (F->taille == 1)
        F->debut = F->fin = NULL;
    else
        F->debut= F->debut->suivant;
   F->taille -= 1;
}

int take_first(fifo *F){
    if (is_empty_fifo(F) == 1){
        printf("La file est vide. Pas de première valeur.");
        return 0;
    }
    else
        return F->debut->valeur;
}

void affiche_fifo(fifo *F) {
   element *e;
   if (F == NULL)
        printf("Erreur; pas d'affichage possible. \n");
   e = F->debut;
   while (e != NULL){
      printf("%d ", e->valeur);
      e = e->suivant;
   }
   printf("\n");
}

/* Structure de tas et fonctions utiles */
typedef struct noeudTas{
    int  v;
    int dist;
} noeudTas;

typedef struct Tas{
    int taille;
    int capacite;
    int *pos;
    struct noeudTas **array;
} Tas;


noeudTas * new_noeudTas(int v, int dist){
    noeudTas *node = malloc(sizeof(*node));
    node->v = v;
    node->dist = dist;
    return node;
}

Tas * creer_tas(int n){
    Tas *tas = malloc(sizeof(*tas));
    tas->pos = (int *) malloc(n * sizeof(int));
    tas->taille = 0;
    tas->capacite = n;
    tas->array = malloc(n *sizeof(noeudTas*));
    return tas;
}

void echange_noeud(noeudTas **a, noeudTas **b){
    noeudTas * t = *a;
    *a = *b;
    *b = t;
}


/* Une fonction pour entasser un noeud dans un tas */
void entasser(Tas* tas, int idx){
    int smallest, left, right;
    smallest = idx;
    left = 2 * idx + 1;
    right = 2 * idx + 2;
    if (left < tas->taille && tas->array[left]->dist < tas->array[smallest]->dist)
        smallest = left;
    if (right < tas->taille && tas->array[right]->dist <tas->array[smallest]->dist )
        smallest = right;
    if (smallest != idx)
    {
        noeudTas *smnode = tas->array[smallest];
        noeudTas *idxnode = tas->array[idx];
        tas->pos[smnode->v] = idx;
        tas->pos[idxnode->v] = smallest;
        echange_noeud(&tas->array[smallest], &tas->array[idx]);
        entasser(tas, smallest);
    }
}

int is_empty_tas(struct Tas* tas){
    return tas->taille == 0;
}

/* Fonction pour extraire le minimum d'un tas. */
noeudTas* extract_min(Tas* tas){
    if (tas -> taille == 0)
        return NULL;
    noeudTas* root = tas->array[0];
    noeudTas* lastnode = tas->array[tas->taille - 1];
    tas->array[0] = lastnode;
    tas->pos[root->v] = tas->taille-1;
    tas->pos[lastnode->v] = 0;
    tas->taille -= 1;
    entasser(tas, 0);
    return root;
}

/* Fonction pour diminuer la valeur d'un certain noeud. */
void decrease_value(struct Tas* tas, int v, int dist){
    int i = tas->pos[v];
    tas->array[i]->dist = dist;
    while (i && tas->array[i]->dist < tas->array[(i - 1) / 2]->dist){
        tas->pos[tas->array[i]->v] = (i-1)/2;
        tas->pos[tas->array[(i-1)/2]->v] = i;
        echange_noeud(&tas->array[i], &tas->array[(i - 1) / 2]);
        i = (i - 1) / 2;
    }
}

/* Fonction pour tester si un noeud est dans un tas. */
int is_in_tas(Tas *tas, int v){
    if (tas->pos[v] < tas->taille)
        return 1;
    return 0;
}

/* Structure de liste d'adjacence */
typedef struct noeudListeAdj{
    int but;
    int poids;
    struct noeudListeAdj* suivant;
} noeudListeAdj;


typedef struct ListeAdj{
   struct noeudListeAdj* head;
} ListeAdj;

int is_empty_adjlist(ListeAdj L){
    return L.head == NULL;
}

typedef struct graphe{
    int vertices;
    struct ListeAdj* array;
} graphe;


/* Crée un noeud de liste d'adjacence */
noeudListeAdj * new_noeudListeAdj(int but, int poids){
    noeudListeAdj *noeud = malloc(sizeof(*noeud));
    noeud->but = but;
    noeud->poids = poids;
    noeud->suivant = NULL;
    return noeud;
}

/* Fonction qui crée un graphe vide à n sommets. */
graphe * new_graph(int n){
    graphe *grph = malloc(sizeof(*grph));
    grph->vertices = n;
    grph->array = (ListeAdj*) malloc(n * sizeof(struct ListeAdj));
    for (int i = 0; i < n; ++i)
        grph->array[i].head = NULL;
    return grph;
}

void add_edge(graphe *grph, int src, int but, int poids){
    if (poids == 0)
        return;
    noeudListeAdj *noeud = new_noeudListeAdj(but, poids);
    noeud->suivant = grph->array[src].head;
    grph->array[src].head = noeud;
}

/* Fonction qui teste si il y a une arête de i vers j dans le graphe,
renvoie le poids de l'arête si oui et 0 sinon. */
int is_edge_in_adjlist(graphe *grph,int i,int j){
    int test = 0;
    noeudListeAdj *n = grph->array[i].head;
    while (n != NULL && test == 0){
        if (n->but == j)
            test = n->poids;
        n=n->suivant;
    }
    return test;
}

/* Affichage du tableau des accessibles */
void print_acc(int *tab, int taille, int s){
    int i;
    printf("Sommets : \t \t");
    for (i=0; i < taille; i++){
        printf("%d",i);
        printf(i<taille-1?"\t":"\n");
    }
    printf("Acc. depuis %d : \t",s);
    for (i=0; i< taille; i++){
        printf("%d",tab[i]);
        printf(i<taille-1?"\t":"\n");
    }
}

/* Affichage du tableau des distances */
void print_dist(int *tab, int taille, int s){
    int i;
    printf("Sommets : \t \t");
    for (i=0; i < taille; i++){
        printf("%d",i);
        printf(i<taille-1?"\t":"\n");
    }
    printf("Dist. depuis %d : \t",s);
    for (i=0; i< taille; i++){
        printf("%d",tab[i]);
        printf(i<taille-1?"\t":"\n");
    }
}

/* Affichage des listes d'adjacence d'un graphe */
void print_adjlist(struct graphe* grph){
    int v;
    for (v = 0; v < grph->vertices; v++) {
        noeudListeAdj* temp = grph->array[v].head;
        printf("\n Liste d'adjacence de %d : \n head ", v);
        while (temp != NULL){
            printf("-> (%d, %d)", temp->but, temp->poids);
            temp = temp->suivant;
        }
        printf("\n");
    }
}

/* Génération d'un graphe aléatoire, avec un certain taux de 0. */
graphe * remplir_graphe(int n){
    graphe *grph = new_graph(n);
    int i,j,w,num,max;
    float v, taux;
    max = n * n;
    taux = 25.0;
    num = n / 10;
    while (num > 1){
        num /= 5;
        taux /= 3.0;
    }
    taux /= 105.0;
    srand(1);
    for (i=0; i < n; i++){
        for (j=0; j < n; j++){
            v = (float) rand () / RAND_MAX;
            add_edge(grph,i,j,v < taux ? (int) (v * 100.) : 0);
        }
    }
    return grph;
}

/* Parcours en largeur avec une file */
void accessibles_list(struct graphe* grph, int s){
    int * vus = malloc( grph->vertices * sizeof(int));
    int k;
    int n=grph->vertices;
    for (k=0; k< grph->vertices; k++)
        vus[k] = 0;
    fifo *aVisiter = new_fifo();
    vus[s] = 1;
    push(aVisiter,s);
    while (aVisiter->taille != 0){
        int u = take_first(aVisiter);
        noeudListeAdj *temp = grph->array[u].head;
        while (temp != NULL){
            int v = temp->but;
            if (vus[v] == 0){
                vus[v] = 1;
                push(aVisiter, v);
            }
            temp = temp->suivant;
        }
        defiler(aVisiter);
    }
    print_acc(vus,n,s);
    free(aVisiter);
    free(vus);
}

/* Dijkstra, version listes d'adjacence, sans file de priorité */
int minDistance(graphe *grph, int dist[], int access[])
{
    int n = grph->vertices;
    int min = INF, min_index;
    int i;
    for (int i = 0; i < n; i++)
        if (access[i] == 0 && dist[i] <= min)
            min = dist[i], min_index = i;
    return min_index;
}

/* Algorithme de Dijkstra, version liste adjacence, sans file de priorité */
void dijkstra_list(graphe *grph, int src){
    int n = grph->vertices;
    int * dist = malloc(n * sizeof(int));
    int * access = malloc(n * sizeof(int));
    int i,j,count;
    for (i=0; i<n; i++){
        dist[i] = INF;
        access[i] = 0;
    }
    dist[src] = 0;
    for (int count = 0; count < grph->vertices - 1; count++) {
        int u = minDistance(grph,dist, access);
        access[u] = 1;
        for (j = 0; j < n; j++){
            if (access[j] == 0 &&
                is_edge_in_adjlist(grph,u,j) != 0 &&
                dist[u] + is_edge_in_adjlist(grph,u,j) < dist[j]) 
                {
                    dist[j] = dist[u] + is_edge_in_adjlist(grph,u,j);
                }
        }
    }
    print_dist(dist,n,src);
    free(access);
    free(dist);
}


/* Dijkstra, version listes d'adjacence, avec une file de priorité */
void dijkstra_list_prio(graphe* grph, int src)
{
    int n = grph->vertices;
    int *dist = malloc(n * sizeof(int));
    Tas* tas = creer_tas(n);
    int i;
    for (i = 0; i < n; i++){
        dist[i] = 9999;
        tas->array[i] = new_noeudTas(i, dist[i]);
        tas->pos[i] = i;
    }
    tas->array[src] = new_noeudTas(src, dist[src]);
    tas->pos[src]=src;
    dist[src] = 0;
    decrease_value(tas, src, dist[src]);
    tas->taille = n;
    while (!is_empty_tas(tas)){
        noeudTas* noeud = extract_min(tas);
        int u = noeud->v;
        noeudListeAdj* vois = grph->array[u].head;
        while (vois != NULL){
            int v = vois->but;
            if (is_in_tas(tas, v) &&
                      dist[u] != 9999 &&
              vois->poids + dist[u] < dist[v])
            {
                dist[v] = dist[u] + vois->poids;
                decrease_value(tas, v, dist[v]);
            }
            vois = vois->suivant;
        }
    }
    print_dist(dist,n,src);
    free(dist);
    free(tas);
}



int main(){
    graphe *grph = remplir_graphe(10);
    print_adjlist(grph);
    printf("\n \n");
    printf("Affichage du résultat avec Dijkstra : \n");
    dijkstra_list(grph,0);
    printf("\n \n");
    printf("Affichage du résultat avec Dijkstra prio: \n");
    dijkstra_list_prio(grph,0);
    free(grph);
    return 0;
/*
    int n=10;
    clock_t td, ta;
    while (n<401){
        graphe *grph=remplir_graphe(n);
        // Test FW
        td = clock();
        for (int i=0;i<n;i++){
            dijkstra_list(grph,i);
        }
        ta=clock();
        printf ("Dijkstra : %d \t \t",(int) (ta-td));
        // Test Dijkstra
        td = clock();
        for (int i=0;i<n;i++){
            dijkstra_list_prio(grph,i);
        }
        ta=clock();
        printf ("Dijkstra avec prio : %d \n",(int) (ta-td));
        n+=50;
    }
*/
}
