#include <stdio.h>
#include <stdlib.h>
#include<time.h>



#define MAX_NOEUD 6
#define INF 9999

/* Structure de file et fonctions utiles */
typedef struct element {
    int valeur;
    struct element *suivant;
} element;


typedef struct fifo{
  element *debut;
  element *fin;
  int taille;
} fifo;

int is_empty_fifo(fifo *F){
    if (F->taille == 0 && F->debut == NULL && F->fin == NULL)
        return 1;
    else
        return 0;
}

fifo * new_fifo(){
   fifo *F = malloc(sizeof(*F));
   if (F == NULL)
        printf("Erreur d'allocation.");
   F->taille = 0;
   F->debut = NULL;
   F->fin = NULL;
   return F;
}

void push(fifo *F, int s){
    if (F == NULL)
        printf ("La file n'existe pas - On ne peut empiler.");
    element *e = malloc(sizeof(*e));
    if (e == NULL)
        printf("Erreur d'allocation");
    e->valeur = s;
    e->suivant = NULL;
    if (F->taille == 0){
        F->debut = e;
        F->fin = e;
    }
    else{
        F->fin->suivant = e;
        F->fin = e;
    }
    F->taille+=1;
}

void defiler(fifo *F) {
    element *e = malloc(sizeof(*e));
    if (F == NULL || F->taille == 0)
        printf("Erreur dépilement.");
    e = F->debut;
    if (F->taille == 1)
        F->debut = F->fin = NULL;
    else
        F->debut= F->debut->suivant;
   F->taille -= 1;
}

int take_first(fifo *F){
    if (is_empty_fifo(F) == 1){
        printf("La file est vide. Pas de première valeur.");
        return 0;
    }
    else
        return F->debut->valeur;
}

void affiche_fifo(fifo *F) {
   element *e;
   if (F == NULL)
        printf("Erreur; pas d'affichage possible. \n");
   e = F->debut;
   while (e != NULL){
      printf("%d ", e->valeur);
      e = e->suivant;
   }
   printf("\n");
}

/* Structure de tas et fonctions utiles */
typedef struct noeudTas{
    int  v;
    int dist;
} noeudTas;

typedef struct Tas{
    int taille;
    int capacite;
    int *pos;
    struct noeudTas **array;
} Tas;


noeudTas * new_noeudTas(int v, int dist){
    noeudTas *node = malloc(sizeof(*node));
    node->v = v;
    node->dist = dist;
    return node;
}

Tas * creer_tas(int n){
    Tas *tas = malloc(sizeof(*tas));
    tas->pos = (int *) malloc(n * sizeof(int));
    tas->taille = 0;
    tas->capacite = n;
    tas->array = malloc(n *sizeof(noeudTas*));
    return tas;
}

void echange_noeud(noeudTas **a, noeudTas **b){
    noeudTas * t = *a;
    *a = *b;
    *b = t;
}


/* Une fonction pour entasser un noeud dans un tas */
void entasser(Tas* tas, int idx){
    int smallest, left, right;
    smallest = idx;
    left = 2 * idx + 1;
    right = 2 * idx + 2;
    if (left < tas->taille && tas->array[left]->dist < tas->array[smallest]->dist)
        smallest = left;
    if (right < tas->taille && tas->array[right]->dist <tas->array[smallest]->dist )
        smallest = right;
    if (smallest != idx){
        noeudTas *smnode = tas->array[smallest];
        noeudTas *idxnode = tas->array[idx];
        tas->pos[smnode->v] = idx;
        tas->pos[idxnode->v] = smallest;
        echange_noeud(&tas->array[smallest], &tas->array[idx]);
        entasser(tas, smallest);
    }
}

int is_empty_tas(struct Tas* tas){
    return tas->taille == 0;
}

/* Fonction pour extraire le minimum d'un tas. */
noeudTas* extract_min(Tas* tas){
    if (tas -> taille == 0)
        return NULL;
    noeudTas* root = tas->array[0];
    noeudTas* lastnode = tas->array[tas->taille - 1];
    tas->array[0] = lastnode;
    tas->pos[root->v] = tas->taille-1;
    tas->pos[lastnode->v] = 0;
    tas->taille -= 1;
    entasser(tas, 0);
    return root;
}

/* Fonction pour diminuer la valeur d'un certain noeud. */
void decrease_value(struct Tas* tas, int v, int dist){
    int i = tas->pos[v];
    tas->array[i]->dist = dist;
    while (i && tas->array[i]->dist < tas->array[(i - 1) / 2]->dist){
        tas->pos[tas->array[i]->v] = (i-1)/2;
        tas->pos[tas->array[(i-1)/2]->v] = i;
        echange_noeud(&tas->array[i], &tas->array[(i - 1) / 2]);
        i = (i - 1) / 2;
    }
}

/* Fonction pour tester si un noeud est dans un tas. */
int is_in_tas(Tas *tas, int v){
    if (tas->pos[v] < tas->taille)
        return 1;
    return 0;
}

/* Affichage du tableau des accessibles */
void print_acc(int *tab, int taille, int s){
    int i;
    printf("Sommets : \t \t");
    for (i=0; i < taille; i++){
        printf("%d",i);
        printf(i<taille-1?"\t":"\n");
    }
    printf("Acc. depuis %d : \t",s);
    for (i=0; i< taille; i++){
        printf("%d",tab[i]);
        printf(i<taille-1?"\t":"\n");
    }
}

/* Affichage du tableau des distances */
void print_dist(int *tab, int taille, int s){
    int i;
    printf("Sommets : \t \t");
    for (i=0; i < taille; i++){
        printf("%d",i);
        printf(i<taille-1?"\t":"\n");
    }
    printf("Dist. depuis %d : \t",s);
    for (i=0; i< taille; i++){
        printf("%d",tab[i]);
        printf(i<taille-1?"\t":"\n");
    }
}


// Structure de graphe avec matrice d'adjacence //
typedef struct graphe{
    int vertices;
    int **adj;
}graphe;


graphe * new_graph(int n){
    int i,j;
    graphe *grph = malloc(sizeof(*grph));
    grph->vertices = n;
    grph->adj= malloc(sizeof(int*) * n);
    for (i=0; i<n; i++)
    {
        grph->adj[i] = malloc(n* sizeof(int));
    }
    for (i=0; i < n; i++){
        for (j=0; j < n; j++){
           grph->adj[i][j] = 0;
        }
    }
    return grph;
}

void * delete_graph(graphe *grph){
    int n=grph->vertices;
    for(int i = 0;i < n;i++)
    {
        free(grph->adj[i]);
    }
    free(grph->adj);
    free(grph);
}

void print_graphe(graphe *grph){
    printf("Nombre de sommets : %d. \n",grph->vertices);
    printf("Matrice d'adjacence : \n");
    int i,j;
    int n=grph->vertices;
    for(i=0; i<n; i++){
        for(j=0; j<n; j++){
            printf("%d",grph->adj[i][j]);
            printf(j<n-1?"\t":"\n");
            }
        }
    }

void add_edge_weight(graphe *grph, int i, int j, int weight){
    grph->adj[i][j] += weight;
}

void add_edge(graphe *grph, int i, int j){
    grph->adj[i][j] += 1;
}

/* Fonction permettant de générer un graphe aléatoire, avec un certain taux de zéros. */
graphe * remplir_graphe(int n){
    graphe *grph = new_graph(n);
    int i,j,w,num,max;
    float v, taux;
    max = n * n;
    taux = 25.0;
    num = n / 10;
    while (num > 1) {
        num /= 5;
        taux /= 3.0;
  }
    taux /= 100.0;
    srand(3);
    for (i=0; i < n; i++){
        for (j=0; j < n; j++){
            v = (float) rand () / RAND_MAX;
            grph->adj[i][j] = v < taux ? (int) (v * 100.) : 0;
        }
    }
    return grph;
    }

/* Renvoie quel sommets sont accessibles depuis le sommet s via un parcours en largeur */
void accessibles_mat(graphe *grph, int src){
    int * vus = malloc( grph->vertices * sizeof(int));
    int k;
    for (k=0; k< grph->vertices; k++)
        vus[k] = 0;
    fifo *aVisiter = new_fifo();
    vus[src] = 1;
    push(aVisiter,src);
    while (aVisiter->taille != 0){
        int i = take_first(aVisiter);
        for (int j=0; j< grph->vertices; j++)
        {
            if (grph->adj[i][j] != 0 && vus[j] == 0){
                vus[j] = 1;
                push(aVisiter,j);
            }
        }
        defiler(aVisiter);
    }
    print_acc(vus,grph->vertices,src);
    free(aVisiter);
    free(vus);
}

/* Fonction permettant de trouver le sommet à plus petite distance parmis les sommets pas encore parcourus,
sans utiliser de file de priorité */
int minDistance(graphe *grph, int dist[], int access[]){
    int n = grph->vertices;
    int min = INF, min_index;
    int i;
    for (int i = 0; i < n; i++){
        if (access[i] == 0 && dist[i] <= min)
            min = dist[i], min_index = i;
    }
    return min_index;
}

/* Algorithme de Dijkstra, version matricielle; affiche la plus courte-distance du sommet src aux autres sommets. */
void dijkstra_mat(graphe *grph, int src){
    int n = grph->vertices;
    int * dist = malloc(n * sizeof(int));
    int * access = malloc(n * sizeof(int));
    int i,j,count;
    for (i=0; i<n; i++){
        dist[i] = INF;
        access[i] = 0;
    }
    dist[src] = 0;
    for (int k = 0; k<n-1; k++){
        int u = minDistance(grph,dist, access);
        access[u] = 1;
        for (j = 0; j < n; j++){
            if (access[j] == 0 &&
                grph->adj[u][j] != 0 &&
                dist[u] + grph->adj[u][j] < dist[j]) 
                {
                    dist[j] = dist[u] + grph->adj[u][j];
                }
        }
    }
    print_dist(dist,n,src);
    free(dist);
    free(access);
}

/* Algorithme de Dijkstra, avec file de priorité implémentée par un tas binaire */
void dijkstra_mat_prio(graphe* grph, int src){
    int n = grph->vertices;
    int *dist = malloc(n * sizeof(int));
    Tas* tas = creer_tas(n);
    int i,v;
    for (i = 0; i < n; i++){
        dist[i] = 9999;
        tas->array[i] = new_noeudTas(i, dist[i]);
        tas->pos[i] = i;
    }
    tas->array[src] = new_noeudTas(src, dist[src]);
    tas->pos[src]=src;
    dist[src] = 0;
    decrease_value(tas, src, dist[src]);
    tas->taille = n;
    while (!is_empty_tas(tas)){
        noeudTas* noeud = extract_min(tas);
        int u = noeud->v;
        for (v=0;v<n;v++){
            if (grph->adj[u][v] != 0 &&
                 is_in_tas(tas,v) &&
                 grph->adj[u][v] + dist[u] < dist[v]){
                    dist[v] = dist[u] + grph->adj[u][v];
                    decrease_value(tas,v,dist[v]); 
                } 
        }
    }
    print_dist(dist, n,src);
    free(dist);
    free(tas);
}

/* Algorithme de Floyd-Marshall, avec initialisation de la matrice de départ à partir de la matrice d'adjacence */


int main(){
    graphe *grph=new_graph(5);
    add_edge_weight(grph,0,4,3);
    add_edge_weight(grph,0,2,18);
    add_edge_weight(grph,1,0,8);
    add_edge_weight(grph,2,1,4);
    add_edge_weight(grph,3,1,1);
    add_edge_weight(grph,4,3,2);
    add_edge_weight(grph,4,1,10);
    print_graphe(grph);
    printf("\n");
    printf("Affichage du résultat avec Floyd-Warshall : \n");
    floydwarshall(grph,0);
    printf("\n \n");
    printf("Affichage du résultat avec Dijkstra : \n");
    dijkstra_mat(grph,0);
    printf("\n \n");
    printf("Affichage du résultat avec Dijkstra prio: \n");
    dijkstra_mat_prio(grph,0);
    delete_graph(grph);
    /*
    int n=10;
    clock_t td, ta;
    while (n<201){
        graphe *grph=remplir_graphe(n);
        // Test FW
        td = clock();
        for (int i=0;i<n;i++)
        {
            floydwarshall(grph,i);
        }
        ta=clock();
        printf ("FW : %d \t \t",(int) (ta-td));
        // Test Dijkstra
        td = clock();
        for (int i=0;i<n;i++)
        {
            dijkstra_mat(grph,i);
        }
        ta=clock();
        printf ("Dijkstra : %d \t \t",(int) (ta-td));
        // Test Dijkstra avec file de priorité
        td = clock();
        for (int i=0;i<n;i++)
        {
            dijkstra_mat_prio(grph,i);
        }
        ta=clock();
        printf ("Dijkstra prio : %d \n",(int) (ta-td));
        n+=50;
    }
    */
    return 0;
}
